from __future__ import annotations

import os
import pty
import shutil
import subprocess
import tempfile
from dataclasses import dataclass

import ROOT
from rich.console import Console
from rich.table import Table

from .types import Any, DetectorOption, Path, PathLike, ShowerOption

_ = ROOT.gSystem.Load("libDelphes")  # type: ignore


class Madgraph5:
    """Wrapper class for Madgraph5.

    It uses commands from MadEvent (`launch -i` in Madgraph5 CLI) to launch runs
    sequentially. It also provides a summary of the runs similar to the website
    provided by Madgraph5.

    Parameters
    ----------
    processes : list[str]
        List of processes to generate.
    executable : PathLike, optional
        Path to the Madgraph5 executable, by default "mg5_aMC".
    model : PathLike, optional
        Path to the model or the name of the model, by default "sm".
    definitions : dict[str, Any], optional
        Dictionary of definitions to pass to Madgraph5, by default {}.
    output : PathLike, optional
        Path to the output directory, by default "madevent".
    log_dir : PathLike, optional
        Path to the log directory relative to the output, by default "Logs".

    The relation between the parameters and the Madgraph5 commands is as follows:

    | Parameters  | Commands or options              |
    | ----------- | -------------------------------- |
    | model       | import model command             |
    | definitions | define command                   |
    | processes   | generate & add process commands  |
    | output      | output command                   |
    """

    def __init__(
        self,
        processes: list[str],
        executable: PathLike = "mg5_aMC",
        model: PathLike = "sm",
        definitions: dict[str, Any] = {},
        output: PathLike = "madevent",
        log_dir: PathLike = "Logs",
    ) -> None:
        # Set properties ----------------------------------------------------- #
        self.executable = executable
        self.model = model
        self.definitions = definitions
        self.processes = processes
        self.output = Path(output)

        self.commands = {
            "pre": [
                *[f"import model {self.model}"],
                *[f"define {k} = {v}" for k, v in self.definitions.items()],
                *[f"generate {self.processes[0]}"],
                *[f"add process {p}" for p in self.processes[1:]],
                *[f"output {self.output}"],
            ],
        }

        # Create output directory -------------------------------------------- #
        command_file = self._cmds_to_file(self.commands["pre"])

        if not self.output.exists():
            # Run Madgraph5
            process = subprocess.Popen(
                f"{self.executable} {command_file}",
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            # Remove py.py file after initialization
            if process.poll() is not None:
                if Path("py.py").exists():
                    Path("py.py").unlink()

            # Check if there's an error message in stderr
            # Note:
            # when importing a model like heft that needs to be downloaded,
            # the success message is wrongly printed in stderr. This is why we
            # check for the word "error" in stderr.
            _, stderr = process.communicate()
            if "error" in stderr.decode().lower():
                # if stderr:
                raise RuntimeError(stderr.decode())

        # After output directory is created, set log directory
        self.log_dir = log_dir

    def launch(
        self,
        shower: ShowerOption = "off",
        detector: DetectorOption = "off",
        settings: dict[str, Any] = {},
        cards: list[PathLike] = [],
        multi_run: int = 1,
        verbose: int = 1,
    ):
        """The launch command in Madgraph5 CLI.

        Parameters
        ----------
        shower : ShowerOption, optional
            Shower tool to use, currently supports "off" and "pythia8", by
            default "off".
        detector : DetectorOption, optional
            Detector tool to use, currently supports "off" and "delphes", by
            default "off".
        settings : dict[str, Any], optional
            Dictionary of settings to pass to Madgraph5, by default {}.
        cards : list[PathLike], optional
            List of cards to pass to Madgraph5, by default [].
        multi_run : int, optional
            Number of runs to launch, by default 1.
        verbose : int, optional
            Verbosity level, 0 for quiet, 1 for simple, by default 1.

        The relation between the parameters and the Madgraph5 commands is as
        follows:

        | shower      | shower option                                     |
        | detector    | detector option                                   |
        | settings    | set command                                       |
        | cards       | card paths directly passed when configuring cards |
        """

        if shower not in ["off", "pythia8"]:
            raise ValueError(f"Unknown shower tool {shower}")
        if detector not in ["off", "delphes"]:
            raise ValueError(f"Unknown detector tool {detector}")

        self.shower = shower
        self.detector = detector
        self.settings = settings
        self.cards = [Path(card).resolve() for card in cards]
        self.multi_run = multi_run

        # -------------------------------------------------------------------- #
        commands = [
            *[f"launch -i {self.output}"],
            *[f"multi_run {self.multi_run}"],
            *[f"shower={self.shower}"],
            *[f"detector={self.detector}"],
            *[f"set {k} {v}" for k, v in self.settings.items()],
            *[f"{card}" for card in self.cards],
        ]
        command_file = self._cmds_to_file(commands)

        # -------------------------------------------------------------------- #
        n_runs = len(list((self.output / "Events").glob("*_banner.txt")))
        run_name = f"run_{n_runs+1:02d}"
        log_file = self.log_dir / f"{run_name}.log"
        self.commands[run_name] = commands

        # -------------------------------------------------------------------- #
        with open(log_file, "w") as f:
            _, child = pty.openpty()
            process = subprocess.Popen(
                f"{self.executable} {command_file}",
                shell=True,
                stdin=child,
                stdout=f,
                stderr=child,
            )

        # -------------------------------------------------------------------- #
        self._check_status(log_file, process, verbose)
        os.close(child)

        # -------------------------------------------------------------------- #
        _, stderr = process.communicate()
        if stderr:
            raise RuntimeError(stderr.decode())  # pragma: no cover

        # -------------------------------------------------------------------- #
        return Madgraph5MultiRun.from_name(run_name, self.output)

    @property
    def executable(self) -> Path:
        return self._executable

    @executable.setter
    def executable(self, value: PathLike):
        if (_executable := shutil.which(value)) is None:
            raise EnvironmentError(f"{value} is not a valid executable")

        self._executable = Path(_executable).resolve()

    @property
    def model(self) -> PathLike:
        return self._model

    @model.setter
    def model(self, value: PathLike):
        self._model = value

    @property
    def definitions(self) -> dict[str, Any]:
        return self._definitions

    @definitions.setter
    def definitions(self, value: dict[str, Any]):
        self._definitions = value

    @property
    def processes(self) -> list[str]:
        return self._processes

    @processes.setter
    def processes(self, value: list[str]):
        self._processes = value

    @property
    def output(self) -> Path:
        return self._output

    @output.setter
    def output(self, value: PathLike):
        _output = Path(value).resolve()
        self._output = _output

    @property
    def log_dir(self) -> Path:
        return self._log_dir

    @log_dir.setter
    def log_dir(self, value: PathLike):
        _log_dir = self.output / value
        _log_dir.mkdir(exist_ok=True)
        self._log_dir = _log_dir

    @property
    def runs(self) -> list[Madgraph5MultiRun]:
        """List of runs in the output directory."""
        run_paths = []
        for i in self.output.glob("Events/run_*"):
            if i.is_dir() and i.name.count("_") == 1:
                run_paths.append(i)

        # Sort the runs by their number
        run_paths = sorted(run_paths, key=lambda x: int(x.name.split("_")[-1]))
        runs = [Madgraph5MultiRun.from_name(i.name, self.output) for i in run_paths]

        return runs

    @classmethod
    def from_output(cls, output: PathLike, executable: PathLike = "mg5_aMC"):
        """Create a Madgraph5 instance from an existing output directory."""
        output = Path(output).resolve()
        if not output.exists():
            raise FileNotFoundError(f"{output} does not exist.")

        proc_card = output / "Cards/proc_card_mg5.dat"

        model = "sm"
        definitions = {}
        processes = []

        with proc_card.open() as f:
            for line in f.readlines():
                if line.startswith("import model"):
                    model = line.split()[2]
                if line.startswith("define"):
                    key, value = line.replace("define ", "").split("=")
                    definitions[key.strip()] = value.strip()

                if line.startswith("generate"):
                    processes.append(line.replace("generate ", "").strip())

                if line.startswith("add process"):
                    processes.append(line.replace("add process ", "").strip())

        return cls(
            processes=processes,
            executable=executable,
            model=model,
            definitions=definitions,
            output=output,
        )

    def summary(self):
        """Print a summary of the runs in the output directory."""
        console = Console()
        table = Table(
            title="\n".join(self.processes),
            caption=f"Output: {self.output}",
        )

        table.add_column("#", justify="right")
        table.add_column("Name")
        table.add_column("Tag")
        table.add_column("Cross section (pb)", justify="center")
        table.add_column("N events", justify="right")
        table.add_column("Seed", justify="right")

        for i, multi_run in enumerate(self.runs):
            table.add_row(
                f"{i}",
                f"{multi_run.name}[{len(multi_run.runs)}]",
                f"{multi_run.tag}",
                f"{multi_run.cross_section:.3e}",
                f"{multi_run.n_events:,}",
                f"{multi_run.seed}",
            )

        console.print(table)

    def _cmds_to_file(self, cmds: list[str]) -> str:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write("\n".join(cmds))
            temp_file_path = temp_file.name
        return temp_file_path

    def _check_status(
        self,
        log_file: Path,
        process: subprocess.Popen,
        verbose=1,
    ) -> None:
        # -------------------------------------------------------------------- #
        status_markers = {
            "Running Survey": "Running Survey",
            "Running Pythia8": "Running Pythia8",
            "Running Delphes": "Running Delphes",
            "storing files": "Storing files",
        }

        current_run = 1
        last_position = 0
        last_known_status = -1  # This will help to avoid repeating the same status.

        # Determine the index of the "storing files" status, since it's the
        # repeating endpoint for each run
        storing_files_idx = list(status_markers.keys()).index("storing files")

        # -------------------------------------------------------------------- #
        # Monitor the process and the log file
        while True:
            if Path("py.py").exists():
                Path("py.py").unlink()

            # Check the log file for status updates
            with open(log_file, "r") as f:
                f.seek(last_position)
                content = f.read()
                last_position = f.tell()

                for idx, (marker, status) in enumerate(status_markers.items()):
                    if marker in content and idx > last_known_status:
                        if verbose:
                            print(status)
                        if idx == storing_files_idx:
                            current_run += 1
                            if current_run > self.multi_run:
                                last_known_status = idx
                            else:
                                last_known_status = -1
                            if verbose:
                                print()
                        else:
                            last_known_status = idx

            # Check if process is still running
            if process.poll() is not None:  # Process has finished
                if verbose:
                    print("Done")
                break


@dataclass
class Madgraph5Run:
    """A single run of Madgraph5.

    Single run includes complete information about the run. It is usually
    launced by the `launch` command in Madgraph5 CLI or the `generate_events`
    command in MadEvent CLI.

    Parameters
    ----------
    name : str
        Name of the run, e.g. "run_01", "run_02_0".
    tag : str
        Tag of the run, by default "" (not the same as "tag_1" in Madgraph5).
    directory : PathLike
        Path to the directory of the run.
    seed : int
        Random seed of the run.
    n_events : int
        Number of events in the run parsed from the banner file and the root
        files. The latter shall prevail.
    cross_section : float
        Cross section of the run.
    collider : str
        Collider configurations of the run.
    events : ROOT.TChain
        TChain of the root files in the run.
    """

    name: str
    tag: str
    directory: Path
    seed: int
    n_events: int
    cross_section: float
    collider: str
    events: ROOT.TChain  # type: ignore

    @classmethod
    def from_directory(cls, directory: PathLike):
        directory = Path(directory)
        if not directory.exists():
            raise FileNotFoundError(f"{directory} does not exist.")

        name = directory.name
        if len(banners := list(directory.glob("*_banner.txt"))) == 0:
            raise FileNotFoundError(f"No banner found in {directory}")
        else:
            banner = banners[0]

        lpps = {"1": "p", "2": "e"}
        tag = ""
        n_events = 0
        seed = 0
        lpp1, lpp2 = "", ""
        ebeam1, ebeam2 = 0.0, 0.0
        cross_section = 0.0
        events = ROOT.TChain("Delphes")  # type: ignore
        for root_file in directory.glob("*.root"):
            events.Add(str(root_file))

        with banner.open() as f:
            for line in f.readlines():
                if "run_tag" in line:
                    tag = line.split("=")[0].strip()

                if "nevents" in line:
                    n_events = int(line.split("=")[0].strip())
                    if n_events != events.GetEntries():
                        n_events = events.GetEntries()  # pragma: no cover

                if "iseed" in line:
                    seed = int(line.split("=")[0].strip())

                if "lpp1" in line:
                    lpp1 = lpps.get(line.split("=")[0].strip(), "")

                if "lpp2" in line:
                    lpp2 = lpps.get(line.split("=")[0].strip(), "")

                if "ebeam1" in line:
                    ebeam1 = float(line.split("=")[0].strip())

                if "ebeam2" in line:
                    ebeam2 = float(line.split("=")[0].strip())

                if "Integrated weight (pb)" in line:
                    cross_section = float(line.split()[-1].strip())

        collider = f"{lpp1}{lpp2}: {ebeam1}x{ebeam2}"

        return cls(
            name, tag, directory, seed, n_events, cross_section, collider, events
        )


@dataclass
class Madgraph5MultiRun:
    """A run launched by multi_run of MadEvent.

    A multi-run is a collection of runs launched by the `multi_run` command in
    MadEvent CLI that is launched by the `launch -i` command in Madgraph5 CLI.
    It usually contains one "run_xx" directory, one "run_xx_banner.txt" file,
    and several "run_xx_y" single runs.

    Parameters
    ----------
    name : str
        Name of the run, e.g. "run_01".
    tag : str
        Tag of the run, by default "" (not the same as "tag_1" in Madgraph5).
    seed : int
        Random seed of the run.
    n_events : int
        Number of events in the run parsed from the banner file and the root
        files. The latter shall prevail.
    cross_section : float
        Cross section of the run.
    collider : str
        Collider configurations of the run.
    events : ROOT.TChain
        TChain of the root files in the run.
    runs : list[Madgraph5Run]
        The single runs in the multi-run.
    """

    name: str
    tag: str
    seed: int
    n_events: int
    cross_section: float
    collider: str
    events: ROOT.TChain  # type: ignore
    runs: list[Madgraph5Run]

    @classmethod
    def from_name(cls, name: str, output: PathLike = "madevent"):
        """Create a multi-run from the name of the run and the output directory."""
        output = Path(output)
        if not output.exists():
            raise FileNotFoundError(f"{output} does not exist.")

        events_dir = output / "Events"
        if len(banners := list(events_dir.glob(f"{name}_banner.txt"))) == 0:
            raise FileNotFoundError(f"No run named {name} found in {output}")
        else:
            banner = banners[0]

        runs = [i for i in events_dir.glob(f"{name}_*") if i.is_dir()]
        runs = sorted(runs, key=lambda x: int(x.name.split("_")[-1]))
        runs = [Madgraph5Run.from_directory(run) for run in runs]

        lpps = {"1": "p", "2": "e"}
        tag = ""
        n_events = 0
        seed = 0
        lpp1, lpp2 = "", ""
        ebeam1, ebeam2 = 0.0, 0.0
        cross_section = 0.0
        events = ROOT.TChain("Delphes")  # type: ignore
        for root_file in events_dir.glob(f"{name}_*/*.root"):
            events.Add(str(root_file))

        with banner.open() as f:
            for line in f.readlines():
                if "run_tag" in line:
                    tag = line.split("=")[0].strip()

                if "nevents" in line:
                    n_events = int(line.split("=")[0].strip())
                    if n_events != events.GetEntries() and events.GetEntries() != 0:
                        n_events = events.GetEntries()

                if "iseed" in line:
                    seed = int(line.split("=")[0].strip())

                if "lpp1" in line:
                    lpp1 = lpps.get(line.split("=")[0].strip(), "")

                if "lpp2" in line:
                    lpp2 = lpps.get(line.split("=")[0].strip(), "")

                if "ebeam1" in line:
                    ebeam1 = float(line.split("=")[0].strip())

                if "ebeam2" in line:
                    ebeam2 = float(line.split("=")[0].strip())

                if "Integrated weight (pb)" in line:
                    cross_section = float(line.split()[-1].strip())

        collider = f"{lpp1}{lpp2}: {ebeam1}x{ebeam2}"

        return cls(
            name,
            tag,
            seed,
            n_events,
            cross_section,
            collider,
            events,
            runs,
        )



# Cuts
from .cuts import CutAndCount  # isort: skip

# Trees
from .trees import GradientBoostedDecisionTree  # isort: skip

# Networks
from .networks.mlps import ToyMultilayerPerceptron  # isort: skip

# Utils
from .approach_utils import load_approach  # isort: skip



from __future__ import annotations

import re
from collections import OrderedDict

from .observables import get_observable


class Filter:
    """Pre-select events based on a list of cuts.

    Parameters
    ----------
    cuts : list[str]
        The list of cuts to apply.
    """

    def __init__(self, cuts: list[str]) -> None:
        self.cuts = cuts
        self.stat = OrderedDict({cut: 0 for cut in cuts})

    def read_event(self, event):
        """Read the event from a TTree."""
        self.event = event
        return self

    def passed(self):
        """Check if the event passed the cuts."""

        def _replace_with_value(match):
            """Replace the string name of an observable with its value."""
            observable_name = match.group(0)  # complete match
            value = get_observable(observable_name).read_event(self.event).value

            if value is not None:
                return str(value)
            else:
                return "float('nan')"

        for cut in self.cuts:
            modified_string = re.sub(
                r"\b(?!\d+\b)([\w\d_]+)\.([\w\d_]+)\b", _replace_with_value, cut
            )
            if eval(modified_string) is False:
                self.stat[cut] += 1
                return False

        return True



from pathlib import Path

import dill as pickle
import keras

from ..types import PathLike


def load_approach(filepath: PathLike, **kwargs):
    filepath = Path(filepath)

    if filepath.suffix == ".pickle":
        with open(filepath, "rb") as f:
            return pickle.load(f)
    else:
        return keras.models.load_model(filepath, **kwargs)



from __future__ import annotations

from typing import Protocol

import numpy as np

from .types import Path, PathLike


class Dataset(Protocol):
    """A protocol for datasets used in HML.

    This protocol follows the same style as the `sklearn.utils.Bunch` class and
    is simpified to quickly save and load datasets.
    """

    @property
    def samples(self) -> np.ndarray:
        """The samples (data points) of the dataset."""
        ...

    @property
    def targets(self) -> np.ndarray:
        """The targets (integer labels) of the dataset."""
        ...

    @property
    def description(self) -> str:
        """The description of the dataset."""
        ...

    def save(self, filepath: PathLike, overwrite: bool) -> None:
        """Save the dataset to a .npz file.

        It is required to save `_type` to distinguish between different datasets
        and load them correctly via `load_dataset` method.
        """
        ...

    @classmethod
    def load(cls, filepath: PathLike) -> Dataset:
        """Load the dataset from a .npz file."""
        ...


def load_dataset(filepath: PathLike) -> Dataset:
    """Load a dataset from a .npz file."""
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(f"File {filepath} does not exist")

    data = np.load(filepath)

    if (_type := data.get("_type").item()) is None:
        raise ValueError(
            f"No `_type` key in {filepath}. Cannot load dataset anonymously, "
            f"try to determine the type of dataset and load it directly from "
            f"the corresponding class."
        )

    if _type not in DATASET_CLASSES:
        raise ValueError(
            f"Unknown dataset type {_type}. Available types: "
            f"{', '.join(DATASET_CLASSES.keys())}"
        )

    return DATASET_CLASSES[_type].load(filepath)


class TabularDataset:
    """A tabular dataset.

    This is a class for tabular datasets. Samples are from 1D representation to
    build this 2D representation.

    Parameters
    ----------
    samples : np.ndarray
        The samples (data points) of the dataset.
    targets : np.ndarray
        The targets (integer labels) of the dataset.
    feature_names : list[str]
        The names of the features.
    target_names : list[str]
        The names of the targets.
    description : str
        The description of the dataset.
    """

    def __init__(
        self,
        samples: np.ndarray,
        targets: np.ndarray,
        feature_names: list[str],
        target_names: list[str],
        description: str,
    ) -> None:
        self._samples = samples
        self._targets = targets
        self._feature_names = feature_names
        self._target_names = target_names
        self._description = description

    @property
    def samples(self) -> np.ndarray:
        return self._samples

    @property
    def targets(self) -> np.ndarray:
        return self._targets

    @property
    def feature_names(self) -> list[str]:
        return self._feature_names

    @property
    def target_names(self) -> list[str]:
        return self._target_names

    @property
    def description(self) -> str:
        return self._description

    def save(self, filepath: PathLike, overwrite: bool = True) -> None:
        filepath = Path(filepath)
        if filepath.exists():
            if overwrite:
                filepath.unlink()
            else:
                raise FileExistsError(f"File {filepath} already exists")

        np.savez(
            filepath,
            samples=self.samples,
            targets=self.targets,
            feature_names=self.feature_names,
            target_names=self.target_names,
            description=self.description,
            _type="tabular",
        )

    @classmethod
    def load(cls, filepath: PathLike) -> TabularDataset:
        filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError(f"File {filepath} does not exist")

        data = np.load(filepath)
        return cls(
            samples=data["samples"],
            targets=data["targets"],
            feature_names=data["feature_names"],
            target_names=data["target_names"],
            description=data["description"].item(),
        )


DATASET_CLASSES = {
    "tabular": TabularDataset,
}



from __future__ import annotations

import tensorflow as tf
from keras.metrics import (
    FalsePositives,
    Metric,
    SpecificityAtSensitivity,
    TruePositives,
)


class MaxSignificance(Metric):
    def __init__(
        self,
        thresholds=None,
        class_id=1,
        name="max_significance",
        dtype=None,
    ):
        super().__init__(name, dtype)

        self.class_id = class_id
        self.tp = TruePositives(thresholds=thresholds, dtype=dtype)
        self.fp = FalsePositives(thresholds=thresholds, dtype=dtype)

    def update_state(self, y_true, y_pred, sample_weight=None):  # pragma: no cover
        y_true = tf.convert_to_tensor(y_true)
        y_pred = tf.convert_to_tensor(y_pred)

        if tf.rank(y_true) == 2 and tf.shape(y_true)[-1] > 1:
            y_true = tf.gather(y_true, self.class_id, axis=-1)

        if tf.rank(y_pred) == 2 and tf.shape(y_pred)[-1] > 1:
            y_pred = tf.gather(y_pred, self.class_id, axis=-1)

        self.tp.update_state(y_true, y_pred, sample_weight)
        self.fp.update_state(y_true, y_pred, sample_weight)

    def result(self):  # pragma: no cover
        s = self.tp.result()
        b = self.fp.result()
        significance = s / tf.sqrt(s + b)
        max_significance = tf.reduce_max(significance)
        return max_significance

    def reset_state(self):  # pragma: no cover
        self.tp.reset_state()
        self.fp.reset_state()


class RejectionAtEfficiency(Metric):
    def __init__(
        self,
        efficiency,
        num_thresholds=200,
        class_id=None,
        name=None,
        dtype=None,
    ):
        super().__init__(name, dtype)
        self.specificity_at_sensitivity = SpecificityAtSensitivity(
            sensitivity=efficiency,
            num_thresholds=num_thresholds,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )

    def update_state(self, y_true, y_pred, sample_weight=None):  # pragma: no cover
        y_true = tf.convert_to_tensor(y_true)
        y_pred = tf.convert_to_tensor(y_pred)

        if y_true.shape != y_pred.shape:
            n_classes = y_pred.shape[-1]
            y_true = tf.one_hot(tf.squeeze(y_true), n_classes, dtype=tf.int32)

        self.specificity_at_sensitivity.update_state(y_true, y_pred, sample_weight)

    def result(self):  # pragma: no cover
        fpr = 1 - self.specificity_at_sensitivity.result()
        rejection = 1 / fpr
        return rejection

    def reset_state(self):  # pragma: no cover
        self.specificity_at_sensitivity.reset_state()



import keras
import tensorflow as tf


@keras.saving.register_keras_serializable()
class CutAndCount(keras.Model):
    def __init__(self, n_bins: int = 50, name: str = "cut_and_count"):
        super().__init__(name=name)
        self.n_bins = self.add_weight(
            name="n_bins",
            shape=(),
            dtype=tf.int32,
            initializer=keras.initializers.Constant(n_bins),
            trainable=False,
        )

    def build(self, input_shape):
        self.cuts = self.add_weight(
            name="cuts",
            shape=(input_shape[-1], 2),
            dtype=tf.float32,
            initializer=keras.initializers.Zeros(),
            trainable=False,
        )
        self.cases = self.add_weight(
            name="cases",
            shape=(input_shape[-1],),
            dtype=tf.int32,
            initializer=keras.initializers.Zeros(),
            trainable=False,
        )

    def train_step(self, data):
        x, y = data

        # Loop over all features to get best results for each feature
        # I: (n_samples,), (n_samples, ?)
        # O: (n_features, 3): (cut0, cut1, case)
        results = tf.map_fn(
            lambda x: self._get_result(x, y),
            tf.transpose(x),
            fn_output_signature=tf.float32,
        )

        # Call the model once to create the weights
        _ = self(x)

        cases = tf.cast(results[:, 2], tf.int32)  # type: ignore
        self.cases.assign(cases)
        self.cuts.assign(results[:, :2])  # type: ignore
        y_pred = self(x)

        loss = self.compute_loss(y=y, y_pred=y_pred)
        self.compiled_metrics.update_state(y, y_pred)  # type: ignore

        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred)

        return {m.name: m.result() for m in self.metrics}

    @tf.function
    def _get_result(self, x, y):  # pragma: no cover
        # Find the min and max of the feature to get the bin edges
        # bin_edges: (n_bins+1,)
        x_min, x_max = tf.reduce_min(x), tf.reduce_max(x)
        bin_edges = tf.linspace(x_min, x_max, self.n_bins + 1)

        # Ensure y is a 1d tensor to perform as a boolean mask
        y_ = tf.cond(
            tf.rank(y) > 1,
            lambda: tf.argmax(y, 1, tf.int32),
            lambda: tf.cast(y, tf.int32),
        )

        # Get the histogram of each class
        x0 = tf.gather(x, tf.where(y_ == 0)[:, 0])
        x1 = tf.gather(x, tf.where(y_ == 1)[:, 0])
        hist0 = tf.histogram_fixed_width(x0, [x_min, x_max], self.n_bins)
        hist1 = tf.histogram_fixed_width(x1, [x_min, x_max], self.n_bins)

        # Find the candidate cuts on the boundary of the two classes
        # For the 0 class
        curr_case0 = hist0 > hist1
        next_case0 = tf.roll(curr_case0, shift=-1, axis=0)
        is_on_boundary0 = tf.math.logical_xor(curr_case0[:-1], next_case0[:-1])
        is_not_empty0 = hist0[:-1] > 0
        is_candidate0 = tf.math.logical_and(is_on_boundary0, is_not_empty0)
        candidate_indices0 = tf.where(is_candidate0)[:, 0] + 1
        candidates0 = tf.gather(bin_edges, candidate_indices0)

        # For the 1 class
        curr_case1 = hist1 > hist0
        next_case1 = tf.roll(curr_case1, shift=-1, axis=0)
        is_on_boundary1 = tf.math.logical_xor(curr_case1[:-1], next_case1[:-1])
        is_not_empty1 = hist1[:-1] > 0
        is_candidate1 = tf.math.logical_and(is_on_boundary1, is_not_empty1)
        candidate_indices1 = tf.where(is_candidate1)[:, 0] + 1
        candidates1 = tf.gather(bin_edges, candidate_indices1)

        # Use unique to remove duplicates
        candidates = tf.unique(tf.concat([candidates0, candidates1], axis=0))[0]

        # Add the max if candidate is only one
        candidates = tf.cond(
            tf.size(candidates) == 1,
            lambda: tf.concat([candidates, [x_max]], axis=0),
            lambda: candidates,
        )

        # Pair the candidates
        i, j = tf.meshgrid(candidates, candidates)
        mask = tf.math.less(i, j)
        i, j = tf.boolean_mask(i, mask), tf.boolean_mask(j, mask)
        candidate_pairs = tf.stack([i, j], axis=1)  # (n_pairs, 2)

        # Loop over all pairs to get best cases
        # I: (n_samples,), (n_samples, ?), (n_pairs, 2)
        # O: (n_pairs, 2): (loss, case)
        losses_and_cases = tf.map_fn(
            lambda pair: self._get_min_loss_and_case(x, y, pair),
            candidate_pairs,
            fn_output_signature=tf.float32,
        )

        # Find the minimum loss for all pairs per feature and take the first
        # candidate as cut
        # (n_pairs,) -> ()
        min_loss = tf.reduce_min(losses_and_cases[:, 0])  # type: ignore
        min_index = tf.where(tf.equal(losses_and_cases[:, 0], min_loss))[0, 0]  # type: ignore
        cut, case = candidate_pairs[min_index], losses_and_cases[min_index, 1]  # type: ignore

        # Finally, pack the cut and case into a tensor and return it
        result = tf.stack([cut[0], cut[1], case], axis=0)

        return result

    @tf.function
    def _get_min_loss_and_case(self, x, y, candidate_pair):  # pragma: no cover
        # I: (n_samples,), (n_samples, ?), (2,)
        cut0 = candidate_pair[0]
        cut1 = candidate_pair[1]

        # For each pair, calculate the loss of the four cases
        # Case 0: signal on the left
        on_left = x <= cut0  # (n_samples,)
        y_pred0 = tf.where(on_left, 1.0, 0.0)  # (n_samples,)
        # Turn the predictions into one-hot encoding
        y_pred0 = tf.concat([1 - y_pred0[:, None], y_pred0[:, None]], axis=1)
        loss0 = self.compute_loss(y=y, y_pred=y_pred0)
        self.compiled_loss.reset_state()  # type: ignore

        # Case 1: signal on the right
        on_right = x >= cut0
        y_pred1 = tf.where(on_right, 1.0, 0.0)
        y_pred1 = tf.concat([1 - y_pred1[:, None], y_pred1[:, None]], axis=1)
        loss1 = self.compute_loss(y=y, y_pred=y_pred1)
        self.compiled_loss.reset_state()  # type: ignore

        # Case 2: signal in the middle
        in_middle = tf.math.logical_and(x >= cut0, x <= cut1)
        y_pred2 = tf.where(in_middle, 1.0, 0.0)
        y_pred2 = tf.concat([1 - y_pred2[:, None], y_pred2[:, None]], axis=1)
        loss2 = self.compute_loss(y=y, y_pred=y_pred2)
        self.compiled_loss.reset_state()  # type: ignore

        # Case 3: signal on both sides
        on_both_sides = tf.math.logical_or(x <= cut0, x >= cut1)
        y_pred3 = tf.where(on_both_sides, 1.0, 0.0)
        y_pred3 = tf.concat([1 - y_pred3[:, None], y_pred3[:, None]], axis=1)
        loss3 = self.compute_loss(y=y, y_pred=y_pred3)
        self.compiled_loss.reset_state()  # type: ignore

        losses_and_cases = tf.concat(
            [
                [[loss0, 0]],
                [[loss1, 1]],
                [[loss2, 2]],
                [[loss3, 3]],
            ],
            axis=0,
        )

        # Find the minimum loss and return it with the corresponding case
        min_loss = tf.reduce_min(losses_and_cases[:, 0])  # type: ignore
        min_index = tf.where(tf.equal(losses_and_cases[:, 0], min_loss))[0, 0]  # type: ignore
        return losses_and_cases[min_index]  # (2,)

    def call(self, x):  # pragma: no cover
        left = x <= self.cuts[:, 0]  # (n_samples, n_features)
        right = x >= self.cuts[:, 0]
        middle = tf.logical_and(self.cuts[:, 0] <= x, x <= self.cuts[:, 1])
        both_sides = tf.logical_or(x <= self.cuts[:, 0], x >= self.cuts[:, 1])
        conditions = tf.stack(
            [left, right, middle, both_sides], axis=-1
        )  # (n_samples, n_features, 4)

        feature_indices = tf.range(tf.shape(x)[1])
        cases_indices = tf.cast(self.cases, tf.int32)
        index_pairs = tf.stack(
            [feature_indices, cases_indices], axis=-1
        )  # (n_features, 2)

        y_pred_per_feature = tf.map_fn(
            lambda index_pair: self.get_pred_per_feature(conditions, index_pair),
            index_pairs,
            fn_output_signature=tf.bool,
        )  # (n_features, n_samples)
        y_pred_per_feature = tf.transpose(y_pred_per_feature)  # (n_samples, n_features)
        y_pred = tf.reduce_all(y_pred_per_feature, axis=-1)  # (n_samples,)
        y_pred = tf.cast(y_pred, tf.float32)
        y_pred = tf.stack([1 - y_pred, y_pred], axis=-1)  # type: ignore
        return y_pred

    # conditions: (n_samples, n_features, 4)
    # indices: (n_features, 2)
    def get_pred_per_feature(self, conditions, indices):
        return conditions[:, indices[0], indices[1]]

    def summary(self, line_length=None, **kwargs):
        super().summary(line_length=line_length, **kwargs)
        print(f"n_bins: {self.n_bins.numpy()}")
        print("cuts:")
        for i, (cuts, case) in enumerate(zip(self.cuts, self.cases), 1):
            if case == 0:
                print(f"  #{i}: x <= {cuts[0]:.4f}")
            elif case == 1:
                print(f"  #{i}: x >= {cuts[0]:.4f}")
            elif case == 2:
                print(f"  #{i}: {cuts[0]:.4f} <= x <= {cuts[1]:.4f}")
            else:
                print(f"  #{i}: x <= {cuts[0]:.4f} or x >= {cuts[1]:.4f}")



import keras


class ToyMultilayerPerceptron(keras.Model):
    def __init__(self, n_classes: int = 2, name: str = "toy_multilayer_perceptron"):
        super().__init__(name=name)
        self.n_classes = n_classes
        self.dense_1 = keras.layers.Dense(10, activation="relu")
        self.dense_2 = keras.layers.Dense(10, activation="relu")
        self.dense_3 = keras.layers.Dense(n_classes, activation="softmax")

    def call(self, inputs):
        x = self.dense_1(inputs)
        x = self.dense_2(x)
        x = self.dense_3(x)
        return x



from __future__ import annotations

import numpy as np
import pandas as pd

from .observables import Observable, get_observable


class Set:
    """A set of observables.

    Set is a 1D representation of an event. It contains a list of observables.
    It is usually used as input to approaches like CutAndCount and
    ToyMultilayerPerceptron (MLP).

    Parameters
    ----------
    observables : list[Observable | str]
        A list of observables or their names.
    """

    def __init__(self, observables: list[Observable | str]) -> None:
        self.observables = []
        for obs in observables:
            if isinstance(obs, Observable):
                self.observables.append(obs)
            else:
                self.observables.append(get_observable(obs))

        self.names = [obs.name for obs in self.observables]
        self.values = []

    def read_event(self, event) -> None:
        for obs in self.observables:
            obs.read_event(event)

        self.values.append([obs.value for obs in self.observables])

    def to_numpy(self, dtype=np.float32) -> np.ndarray:
        return np.array(self.values, dtype=dtype)

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(self.values, columns=self.names)



from __future__ import annotations

from abc import ABC, abstractmethod
from itertools import product
from typing import Any

import numpy as np
from ROOT import TTree  # type: ignore


def get_observable(name: str, **kwargs) -> Observable:
    """Get an observable from its name.

    An observable name is composed of two parts: the shortcut and the class name.
    For example, `Jet_0.Pt` is the name of the `Pt` observable of the first jet,
    in which `Jet_0` is the shortcut and `Pt` is the class name.
    """
    if len(parts := name.split(".")) == 1:
        shortcut, classname = "", parts[0]
    else:
        shortcut, classname = parts

    if classname not in Observable.all_observables:
        raise ValueError(f"Observable {classname} not found")

    return Observable.all_observables[classname](shortcut, **kwargs)


class Observable(ABC):
    """Base class for observables.

    Implement the `get_value` method to define a new observable.
    """

    all_observables = {}

    def __init__(
        self,
        shortcut: str | None = None,
        object_pairs: list[tuple[str, int | None]] | None = None,
    ) -> None:
        self.seprate_branch_name_and_index = "_"
        self.separate_objects = "-"

        if shortcut:
            self.shortcut = shortcut
            self.object_pairs = self.parse_shortcut(shortcut)
        elif object_pairs:
            self.shortcut = self.build_shortcut(object_pairs)
            self.object_pairs = object_pairs
        else:
            self.shortcut = None
            self.object_pairs = None

        self._value = None

    @property
    def name(self) -> str:
        """The name of the observable.

        The name is composed of the shortcut and the class name, e.g. `Jet_0.Pt`.
        """
        if self.shortcut:
            return f"{self.shortcut}.{self.__class__.__name__}"
        else:
            return self.__class__.__name__

    @property
    def value(self) -> Any:
        """The value of the observable."""
        return self._value

    def to_numpy(self) -> np.ndarray:
        """Convert the value to a numpy array."""
        return np.array(self.value, dtype=np.float32)

    def __repr__(self) -> str:
        """The representation of the observable."""
        return f"{self.name}: {self.value}"

    def read_event(self, event: TTree) -> Observable:
        """Read an event and fetch the needed physics objects.

        It creates three attributes: event, objects and _value. The _value is
        calculated by calling the `get_value` method.
        """
        self.event = event
        self.objects = []
        self._value = None

        if self.object_pairs:
            for branch_name, index in self.object_pairs:
                if branch_name in event.GetListOfBranches():
                    branch = getattr(event, branch_name)

                    if index is None:
                        self.objects.append([i for i in branch])
                    elif index < branch.GetEntries():
                        self.objects.append(branch[index])

        self._value = self.get_value()

        return self

    @abstractmethod
    def get_value(self) -> Any:
        """Calculate the value of the observable.

        Implement this method to define a new observable. Here're common cases:
        1. Quick calculation: use `self.event` to get physics objects and write
        the calculation directly.
        2. Take the advantage of Observable: use `self.objects` to do the
        calculation.

        Return None if the observable is not correctly got.
        """
        ...  # pragma: no cover

    def parse_shortcut(self, shortcut: str) -> list[tuple[str, int | None]]:
        """Parse the shortcut to get the object pairs."""
        object_pairs = []
        objects = shortcut.split(self.separate_objects)
        for obj in objects:
            if "_" in obj:
                branch_name, index = obj.split(self.seprate_branch_name_and_index)
                index = int(index)
            else:
                branch_name, index = obj, None

            object_pairs.append((branch_name, index))
        return object_pairs

    def build_shortcut(self, object_pairs: list[tuple[str, int | None]]) -> str:
        """Build the shortcut from the object pairs."""
        shortcuts = []
        for branch_name, index in object_pairs:
            if index is not None:
                obj = f"{branch_name}{self.seprate_branch_name_and_index}{index}"
            else:
                obj = branch_name
            shortcuts.append(obj)
        return self.separate_objects.join(shortcuts)

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        Observable.all_observables[cls.__name__] = cls

    @classmethod
    def add_alias(cls, *alias: str) -> None:
        """Add alias for the class name."""
        for i in alias:
            Observable.all_observables[i] = cls


class Px(Observable):
    """Get the x component of the momentum.

    Available for single and multiple objects. For example:
    - `Jet_0.Px` is the x component of the momentum of the leading jet.
    - `Jet.Px` is the x component of the momentum of all jets.

    Alias: px
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().Px() for i in obj]
        else:
            return obj.P4().Px()


class Py(Observable):
    """Get the y component of the momentum.

    Available for single and multiple objects. For example:
    - `Jet_0.Py` is the y component of the momentum of the leading jet.
    - `Jet.Py` is the y component of the momentum of all jets.

    Alias: py
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().Py() for i in obj]
        else:
            return obj.P4().Py()


class Pz(Observable):
    """Get the z component of the momentum.

    Available for single and multiple objects. For example:
    - `Jet_0.Pz` is the z component of the momentum of the leading jet.
    - `Jet.Pz` is the z component of the momentum of all jets.

    Alias: pz
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().Pz() for i in obj]
        else:
            return obj.P4().Pz()


class E(Observable):
    """Get the energy of the object.

    Available for single and multiple objects. For example:
    - `Jet_0.E` is the energy of the leading jet.
    - `Jet.E` is the energy of all jets.

    Alias: e, Energy
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().E() for i in obj]
        else:
            return obj.P4().E()


class Pt(Observable):
    """Get the transverse momentum of the object.

    Available for single and multiple objects. For example:
    - `Jet_0.Pt` is the transverse momentum of the leading jet.
    - `Jet.Pt` is the transverse momentum of all jets.

    Alias: pt, pT, PT
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().Pt() for i in obj]
        else:
            return obj.P4().Pt()


class Eta(Observable):
    """Get the pseudorapidity of the object.

    Available for single and multiple objects. For example:
    - `Jet_0.Eta` is the pseudorapidity of the leading jet.
    - `Jet.Eta` is the pseudorapidity of all jets.

    Alias: eta
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().Eta() for i in obj]
        else:
            return obj.P4().Eta()


class Phi(Observable):
    """Get the azimuthal angle of the object.

    Available for single and multiple objects. For example:
    - `Jet_0.Phi` is the azimuthal angle of the leading jet.
    - `Jet.Phi` is the azimuthal angle of all jets.

    Alias: phi
    """

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().Phi() for i in obj]
        else:
            return obj.P4().Phi()


class M(Observable):
    """Get the mass of the object.

    Available for single and multiple objects. For example:
    - `Jet_0.M` is the mass of the leading jet.
    - `Jet.M` is the mass of all jets.

    Alias: m, mass, Mass"""

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.P4().M() for i in obj]
        else:
            return obj.P4().M()


class DeltaR(Observable):
    """Calculate the deltaR between two objects.

    Available for two objects. For example:
    - `Jet_0-Jet_1.DeltaR` is the deltaR between the leading two jets.
    - `Jet_0-Jet.DeltaR` is the deltaR between the leading jet and all jets.
    - `Jet-Jet.DeltaR` is the deltaR between all jets.

    """

    def get_value(self) -> Any:
        if len(self.objects) != 2:
            return

        obj1, obj2 = self.objects
        obj1 = [obj1] if not isinstance(obj1, list) else obj1
        obj2 = [obj2] if not isinstance(obj2, list) else obj2

        distances = []
        for i, j in product(obj1, obj2):
            distances.append(i.P4().DeltaR(j.P4()))

        return distances[0] if len(distances) == 1 else distances


class NSubjettiness(Observable):
    """Get the n-subjettiness from the leaf Tau of the branch FatJet.

    Available for single FatJet objects. For example:
    - `FatJet_0.NSubjettiness` is the tau1 (by default) of the leading FatJet.

    Alias: TauN
    """

    def __init__(
        self,
        shortcut: str | None = None,
        object_pairs: list[tuple[str, int | None]] | None = None,
        n: int = 1,
    ) -> None:
        super().__init__(shortcut, object_pairs)
        self.n = n

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.Tau[self.n - 1] for i in obj]
        else:
            return obj.Tau[self.n - 1]


class NSubjettinessRatio(Observable):
    """Calculate the n-subjettiness ratio from the leaf Tau of the branch FatJet.

    Available for single FatJet objects. For example:
    - `FatJet_0.NSubjettinessRatio` is the tau21 (by default) of the leading
    FatJet.

    Alias: TauMN
    """

    def __init__(
        self,
        shortcut: str | None = None,
        object_pairs: list[tuple[str, int | None]] | None = None,
        m: int = 2,
        n: int = 1,
    ) -> None:
        super().__init__(shortcut, object_pairs)
        self.m = m
        self.n = n

    def get_value(self) -> Any:
        if len(self.objects) == 0:
            return

        obj = self.objects[0]
        if isinstance(obj, list):
            return [i.Tau[self.m - 1] / i.Tau[self.n - 1] for i in obj]
        else:
            return obj.Tau[self.m - 1] / obj.Tau[self.n - 1]


class Size(Observable):
    """The number of physics objects.

    Available for single object. For example:
    - `Jet.Size` is the number of jets.

    Alias: size
    """

    def get_value(self):
        if len(self.objects) > 0:
            return len(self.objects[0])


Px.add_alias("px")
Py.add_alias("py")
Pz.add_alias("pz")
E.add_alias("e", "Energy")

Pt.add_alias("pt", "pT", "PT")
Eta.add_alias("eta")
Phi.add_alias("phi")
M.add_alias("m", "mass", "Mass")

NSubjettiness.add_alias("TauN")
NSubjettinessRatio.add_alias("TauMN")

Size.add_alias("size")

Energy = E
PT = Pt
Mass = M
TauN = NSubjettiness
TauMN = NSubjettinessRatio
